#include <windows.h>
#include <tchar.h>
#pragma warning(disable:4819)	// regext.h /
#include <regext.h>
#pragma warning(default:4819)

#include "DataSetting.h"
#include "Bg.h"
#include "Bar.h"

BOOL Registry::Open()
{
  HKEY hPKey;

  hPKey = NULL; hKey = NULL;
  if (((RegCreateKeyEx(HKEY_CURRENT_USER,APP_REGISTRY,0,NULL,
		       REG_OPTION_VOLATILE, KEY_ALL_ACCESS, 
		       NULL, &hPKey, NULL))||
       (RegCreateKeyEx(hPKey, lpSubKey,0,NULL,
		       REG_OPTION_VOLATILE, KEY_ALL_ACCESS, 
		       NULL, &hKey, NULL))),
      ((hPKey!=NULL)?RegCloseKey(hPKey):TRUE)) {
    return FALSE;
  }
  return TRUE;
}

BOOL Registry::Close()
{
  RegCloseKey(hKey);
  hKey = NULL;
  return TRUE;
}

BOOL Registry::SaveRaw(const BYTE *pData, DWORD dwSize, int nIdx)
{
  BOOL bResult;
  TCHAR sIdx[3];

  if (!Open()) return FALSE;
  _itot(nIdx, sIdx, 10);
  bResult = RegSetValueEx(hKey, sIdx, NULL, REG_BINARY, pData, dwSize);
  Close();    
  return bResult;
}

Registry::Registry(const struct BarEntry *pBE)
{
  lpSubKey         = pBE->lpRegistry;
  lpDefaultOptions = pBE->lpDefaultOptions;
  hKey             = NULL;

  nDefaultSet      = 0;		// dummy
  for (int i=0;i<2;i++) {
    if (Load(NULL, i)!=0) {  	// Setting Found
      nDefaultSet = i;
      nCurrentSet = -1;
      return;
    }
  }
  nDefaultSet      = -1;
  nCurrentSet      = -1;
}

Registry::~Registry()
{
  Close();
}

int Registry::Select(int nMode, int nIdx)
{
  int nIndex;

  switch (nMode) {
  case 0: // Reset
    nIndex = nDefaultSet;
    break;
  case 1: // Next
    nIndex = nCurrentSet + 1;
    break;
  case 2: // Direct
    nIndex = nIdx;
    break;
  default:// No Operation
    nIndex = nCurrentSet;
    break;
  }
  if (nCurrentSet == nIndex) return -1;
  return nIndex;
}

DWORD Registry::Load(BarSetting *pBS, int nIdx)
{
  BgSetting *pBGS;
  BYTE *p;
  DWORD dwType, dwSize, dwParam;
  struct {
    int nRev;
    int nNum;
    int nHeight;
  } sHeader;

  nCurrentSet = -1;
  if (nDefaultSet != -1) {
    // Get data from Registry
    if ((nIdx == -1)||!Open()) return 0;
    {
      TCHAR sIdx[3];

      _itot(nIdx, sIdx, 10);
      dwType = REG_NONE;
      if ((((RegQueryValueEx(hKey, sIdx, NULL, NULL, NULL, &dwSize)))||
	   ((p = (BYTE *)alloca(dwSize)),
	    (RegQueryValueEx(hKey, sIdx, NULL, &dwType, p, &dwSize)))),
	  Close(),
	  (dwType != REG_BINARY)) {
	return 0;
      }
    }

    // Get Header
    sHeader.nRev = *p++;
    sHeader.nNum = *p++;
    sHeader.nHeight = *p++;
    if (sHeader.nHeight==0) sHeader.nHeight = DEFAULT_HEIGHT;

    // Error check
    if ((sHeader.nRev!=REGVERSION)||
	(GetRegistrySize(sHeader.nNum)>(int)dwSize)){
      return Load(pBS, nDefaultSet);
    }

    // Read Body
    if (pBS!=NULL) {
      pBGS = pBS->BG;
      for (int i=0;i<sHeader.nNum;i++) {
	pBGS->tcOption = (TCHAR)*p++;
	pBGS->bCaps    = *p++;
	pBGS->nHeight  = sHeader.nHeight;
	for (int k=0;k<MAXPARAMNUM;k++) {
	  dwParam = 0;
	  for (int l=0;l<4;l++) dwParam = (dwParam<<8)|*p++;
	  pBGS->nParam[k] = dwParam;
	}
	if (!Bg::CheckOption(pBGS->tcOption)) {
	  sHeader.nNum--;
	  continue;
	}
	pBGS++;
      }
      pBS->nNum    = sHeader.nNum;
      pBS->nHeight = sHeader.nHeight;
    }
  } else {
    // Get data from DefaultOptions
    nIdx = -1;
    sHeader.nNum = _tcslen(lpDefaultOptions);
    sHeader.nHeight = DEFAULT_HEIGHT;
    if (pBS!=NULL) {
      pBGS = pBS->BG;
      for (int i=0;i<sHeader.nNum;i++) {
	pBGS->tcOption = lpDefaultOptions[i];
	pBGS->bCaps    = FALSE;
	pBGS->nHeight  = sHeader.nHeight;
	for (int k=0;k<MAXPARAMNUM;k++) pBGS->nParam[k] = 0;
	pBGS++;
      }
      pBS->nNum    = sHeader.nNum;
      pBS->nHeight = sHeader.nHeight;
    }
  }
  nCurrentSet = nIdx;
  return GetSettingSize(sHeader.nNum);
}

BOOL Registry::Save(BarSetting *pBS, int nIdx)
{
  BYTE *pData = (BYTE *)alloca(GetRegistrySize(pBS->nNum));
  BgSetting *pBGS;
  BYTE *p;

  // make header
  pData[0] = REGVERSION;	// Version
  pData[1] = pBS->nNum;		// nNum
  pData[2] = pBS->nHeight;	// Height

  // make body
  p = &pData[3];		// Body
  pBGS = pBS->BG;
  for (int i=0; i<pBS->nNum; i++) {
    *p++ = (BYTE) pBGS->tcOption;
    *p++ = (BYTE) pBGS->bCaps;
    for (int j=0;j<MAXPARAMNUM;j++)
      for (int k=0;k<4;k++)
	*p++ = (BYTE) ((pBGS->nParam[j]>>((3-k)*8))&0xFFU);
    pBGS++;
  }

  // write registry
  SaveRaw(pData, GetRegistrySize(pBS->nNum), nIdx);
  return TRUE;
}

BOOL Registry::Delete(int nIdx)
{
  BOOL bResult;

  if (nIdx==0) {
    if (RegOpenKeyEx(HKEY_CURRENT_USER, APP_REGISTRY, 0, KEY_READ, &hKey))
      return FALSE;
    bResult = RegDeleteKey(hKey, lpSubKey);
  } else {
    if (Open()) return FALSE;

    TCHAR sIdx[3];

    _itot(nIdx, sIdx, 10);
    bResult = RegDeleteValue(hKey, sIdx);
  }
  Close();
  return bResult;
}

//-----------------------------------------------------------------------------
// R}hC

#define DELIMITER _T(" \t")
#define OPTION	  _T("/-")

static inline int CheckSubOption(const BarEntry *pBE, int nBarNum,
				 LPTSTR *pArg, int nDefault)
{
  int nBar;

  // Check SubOption
  nBar = nDefault;
  for (int i=0; i<nBarNum; i++) {
    if (pBE[i].tcOption == **pArg) {
      (*pArg)++;
      nBar = i;
      break;
    }
  }
  return nBar;
}
static inline UINT CheckSubOptionMulti(const BarEntry *pBE, int nBarNum,
				       LPTSTR *pArg)
{
  UINT bmBar;
  int nBar;

  // Check (Multiple) SubOption
  bmBar = 0;
  for (int i=0; i<nBarNum; i++) {
    if ((nBar=CheckSubOption(pBE, nBarNum, pArg, -1))<0) break;
    bmBar = 1<<nBar;
  }
  if (bmBar == 0) bmBar = 0xffffffffU;
  return bmBar;
}
static inline int GetArgNum(LPTSTR *pArg)
{
  // Get next argument
  if ((**pArg == _T('\0'))&&((*pArg = _tcstok(NULL, DELIMITER))==NULL))
    return 0;
  // Get argument value
  return _tcstol(*pArg, NULL, 0);
}

BOOL Registry::CommandLine(const struct BarEntry *pBE,
			   int nBarNum, int nMaxBGNum,
			   const struct OptionEntry *pOptions, int *nResults, 
			   int nOptions,
			   LPTSTR lpCmdLine, int nIdx)
{
  BarSetting **pBS;
  TCHAR *pArg, nOpt;
  BOOL   bNext, bOptCase;
  int	*pParam, nParam;
  int	 nBar;
  UINT	 bmBar;

  // [N
  pBS = (BarSetting **) alloca(sizeof(BarSetting *) * nBarNum);
  for (int i=0;i<nBarNum;i++) {
    pBS[i] = (BarSetting *) alloca(GetSettingSize(nMaxBGNum));
    pBS[i]->nNum = pBS[i]->nHeight = 0;
  }
  for (int i=0; i<nOptions; i++) nResults[i] = 0;
  pParam = NULL;

  // 
  pArg = _tcstok(lpCmdLine, DELIMITER);
  while (pArg != NULL) {
    if (_tcschr(OPTION, *pArg)!=NULL) {
      pArg++;
      // check global options
      bNext = TRUE;
      for (int i=0; i<nOptions; i++) {
	if (_tcsnicmp(pOptions[i].lpOptions, pArg, pOptions[i].nLen)==0) {
	  pArg += pOptions[i].nLen;

	  bmBar  = 0;
	  nParam = TRUE;
	  if (pOptions[i].bSubOption) 
	    bmBar=CheckSubOptionMulti(pBE,nBarNum,&pArg);
	  if ((pOptions[i].bArg)&&((nParam=GetArgNum(&pArg))==0)) return FALSE;
	  nResults[i]=MAKELONG(nParam,bmBar);
	  bNext=FALSE;
	  break;
	}
      }
      if (bNext) {
	// other options
	nOpt     = _totlower(*pArg);
	bOptCase = _istupper(*pArg);
	pArg++;
	// Check '/h' option
	if (nOpt == _T('h')) {
	  bmBar = CheckSubOptionMulti(pBE, nBarNum, &pArg);
	  if ((nParam = GetArgNum(&pArg))==0) return FALSE;
	  // Set Bar Height
	  for (int i=0; i<nBarNum; i++)
	    if (bmBar&(1<<i)) pBS[i]->nHeight = nParam;
	} else {
	  // Check Option
	  if (! Bg::CheckOption(nOpt)) return FALSE;

	  // Check SubOption
	  nBar = CheckSubOption(pBE, nBarNum, &pArg, 0);
	  if (pBS[nBar]->nNum > nMaxBGNum) return FALSE; // too may options

	  // allocate BGSetting Area
	  BgSetting *pBGS = &pBS[nBar]->BG[pBS[nBar]->nNum++];

	  pBGS->tcOption = nOpt;
	  pBGS->bCaps    = bOptCase;
	  pParam         = pBGS->nParam;

	  // Clear parameter area
	  for (int i=0;i<MAXPARAMNUM; i++) pParam[i] = 0;
	  // Check Augument
	  if (*pArg != _T('\0')) continue;
	}
      }
    } else {
      if (pParam == NULL) return FALSE;
      for (int i=0;i<MAXPARAMNUM;i++) {
	*pParam++ = _tcstol(pArg, &pArg, 0);
	if (*pArg++ != _T(',')) break;
      }
      pParam = NULL;
    }

    pArg = _tcstok(NULL, DELIMITER);
  }

  // Save/Delete registry command
  int nNumAll = 0;
  for (int i=0;i<nBarNum;i++) nNumAll += pBS[i]->nNum;
  for (int i=0;i<nBarNum;i++) {
    Registry pReg(&pBE[i]);

    if (nNumAll != 0) {
      pReg.Save(pBS[i], nIdx);	// R}hCۑ
    } else {
      pReg.Delete(nIdx);
    }
  }
  return TRUE;
}
